@@ -23,7 +23,7 @@
 
 #include "pango-ot-private.h"
 #include "pango-impl-utils.h"
-#include FT_MODULE_H
+#include FT_TRUETYPE_TABLES_H
 
 static void pango_ot_info_class_init (GObjectClass *object_class);
 static void pango_ot_info_finalize   (GObject *object);
@@ -90,6 +90,32 @@ pango_ot_info_finalizer (void *object)
   g_object_unref (info);
 }
 
+static hb_blob_t *
+_get_table  (hb_tag_t tag, void *user_data)
+{
+  PangoOTInfo *info = (PangoOTInfo *) user_data;
+  FT_Byte *buffer;
+  FT_ULong  length = 0;
+  FT_Error error;
+
+  error = FT_Load_Sfnt_Table (info->face, tag, 0, NULL, &length);
+  if (error)
+    return hb_blob_create_empty ();
+
+  buffer = g_malloc (length);
+  if (buffer == NULL)
+    return hb_blob_create_empty ();
+
+  error = FT_Load_Sfnt_Table (info->face, tag, 0, buffer, &length);
+  if (error)
+    return hb_blob_create_empty ();
+
+  return hb_blob_create ((const char *) buffer, length,
+			 HB_MEMORY_MODE_WRITEABLE,
+			 g_free, NULL);
+}
+
+
 /**
  * pango_ot_info_get:
  * @face: a <type>FT_Face</type>.
@@ -110,8 +136,6 @@ pango_ot_info_get (FT_Face face)
     return face->generic.data;
   else
     {
-      hb_blob_t *blob;
-
       if (face->generic.finalizer)
         face->generic.finalizer (face->generic.data);
 
@@ -120,13 +144,21 @@ pango_ot_info_get (FT_Face face)
 
       info->face = face;
 
-      /* XXX handle face->stream->base == NULL better */
-      blob = hb_blob_create ((const char *) face->stream->base,
-			     (unsigned int) face->stream->size,
-			     HB_MEMORY_MODE_READONLY_MAY_MAKE_WRITEABLE,
-			     NULL, NULL);
-      info->hb_face = hb_face_create_for_data (blob, face->face_index);
-      hb_blob_destroy (blob);
+      if (face->stream->base != NULL) {
+	hb_blob_t *blob;
+
+	blob = hb_blob_create ((const char *) face->stream->base,
+			       (unsigned int) face->stream->size,
+			       HB_MEMORY_MODE_READONLY_MAY_MAKE_WRITEABLE,
+			       NULL, NULL);
+	info->hb_face = hb_face_create_for_data (blob, face->face_index);
+	hb_blob_destroy (blob);
+      } else {
+	info->hb_face = hb_face_create_for_tables (_get_table, NULL, info);
+      }
+
+
+      hb_face_set_unicode_funcs (info->hb_face, hb_glib_get_unicode_funcs ());
 
       /* XXX this is such a waste if not SFNT */
       if (!hb_ot_layout_has_font_glyph_classes (info->hb_face))
@@ -423,7 +455,7 @@ pango_ot_info_list_scripts (PangoOTInfo      *info,
 {
   hb_tag_t tt = get_hb_table_type (table_type);
   PangoOTTag *result;
-  unsigned int count;
+  unsigned int count = 0;
 
   hb_ot_layout_table_get_script_tags (info->hb_face, tt, &count, NULL);
   result = g_new (PangoOTTag, count + 1);
@@ -453,7 +485,7 @@ pango_ot_info_list_languages (PangoOTInfo      *info,
 {
   hb_tag_t tt = get_hb_table_type (table_type);
   PangoOTTag *result;
-  unsigned int count;
+  unsigned int count = 0;
 
   hb_ot_layout_script_get_language_tags (info->hb_face, tt, script_index, &count, NULL);
   result = g_new (PangoOTTag, count + 1);
@@ -487,7 +519,7 @@ pango_ot_info_list_features  (PangoOTInfo      *info,
 {
   hb_tag_t tt = get_hb_table_type (table_type);
   PangoOTTag *result;
-  unsigned int count;
+  unsigned int count = 0;
 
   hb_ot_layout_language_get_feature_tags (info->hb_face, tt, script_index, language_index, &count, NULL);
   result = g_new (PangoOTTag, count + 1);
@@ -509,7 +541,7 @@ _pango_ot_info_substitute  (const PangoOTInfo    *info,
       PangoOTRule *rule = &g_array_index (ruleset->rules, PangoOTRule, i);
       hb_mask_t mask;
       unsigned int lookup_count, j;
-      unsigned int lookup_indexes[100];
+      unsigned int lookup_indexes[1000];
 
       if (rule->table_type != PANGO_OT_TABLE_GSUB)
 	continue;
@@ -542,7 +574,7 @@ _pango_ot_info_position    (const PangoOTInfo    *info,
   hb_buffer_clear_positions (buffer->buffer);
 
   /* XXX reuse hb_font */
-  hb_font = hb_font_create (info->hb_face);
+  hb_font = hb_font_create ();
   hb_font_set_scale (hb_font,
 		     info->face->size->metrics.x_scale,
 		     info->face->size->metrics.y_scale);
@@ -556,7 +588,7 @@ _pango_ot_info_position    (const PangoOTInfo    *info,
       PangoOTRule *rule = &g_array_index (ruleset->rules, PangoOTRule, i);
       hb_mask_t mask;
       unsigned int lookup_count, j;
-      unsigned int lookup_indexes[100];
+      unsigned int lookup_indexes[1000];
 
       if (rule->table_type != PANGO_OT_TABLE_GPOS)
 	continue;
